from __future__ import absolute_import, division, print_function
import numpy as np

def KL_divergence_estimator(X, Y, k=1):
    """ Estimate symmetric version of KL divergence. 
    The symmetric version is 0.5 * [ D(P|Q) + D(Q|P) ].
     
    Parameters
    ----------
        X, Y: numpy array
            2-dimensional array where each row is a sample.
        k: int optional
            k-NN to be used. The default is k=1.
    
    return
    ------
    numpy array : estimated D(P|Q)
    """
    if not (isinstance(k, int) or k is None):
        raise ValueError('k has incorrect type.')
    if k is not None and k <= 0:
        raise ValueError('k cannot be <= 0')

    kl1 =  KL_divergence_estimator_sub(X, Y, k=k)
    kl2 =  KL_divergence_estimator_sub(Y, X, k=k)
    
    return (kl1 + kl2) / 2.0


def KL_divergence_estimator_sub(X, Y, k=1):
    """ KL-Divergence estimator universal k-NN estimator.
    
    Parameters
    ----------
        X, Y: numpy array
            2-dimensional array where each row is a sample.
        k: int, optional
            k-NN to be used. The default is k=1.

    return
    ------
    numpy array : estimated D(P|Q)
    """
    n, m = len(X), len(Y)
    D = np.log(m / (n - 1))
    d = float(X.shape[1])

    for xi in X:
        nu = knn_distance(xi, Y, k-1)
        rho = knn_distance(xi, X, k)
        D += (d/n)*np.log(nu/rho)

    return D


def knn_distance(point, sample, k):
    """ Euclidean distance from `point` to it's `k`-Nearest
    Neighbour in `sample`.
    
    Parameters
    ----------
        point: float
            a data point from a sample.
        sample: numpy array
            2-dimensional array where each row is a sample.
        k: int
            k-NN to be used.

    return
    ------
    numpy-array : `k`-Nearest Neighbour in `sample` 
    """
    norms = np.linalg.norm(sample-point, axis=1)
    return np.sort(norms)[k]





from gaussian_vs_lognormal import test_gaussian_vs_lognormal

from cluster_catalog_comp import RM_clusters_consistency_check, RM_clusters_witness_function

from mnist_test import mnist_digits_distance



def mnist_digits_distance():

    import numpy as np
    from sklearn import datasets
    import matplotlib.pyplot as plt
    from tatter import two_sample_test
    from sklearn.metrics import pairwise_distances
    import os

    plot_path = './tatter/tests/plots/'
    if not os.path.exists(plot_path):
        os.makedirs(plot_path)

    digits = datasets.load_digits()
    data = digits.data
    target = digits.target

    num = [data[target == i] for i in range(10)]

    plt.figure(figsize=(16, 16))

    for i in range(1, 9):
        for j in range(8):
            ax = plt.subplot(9, 8, i * 8 + j + 1)

            if i == 1: ax.set_title('%i' % (j + 1), size=25, fontweight="bold")
            if j == 0: ax.set_ylabel('%i' % (i + 1), size=25, fontweight="bold")
            ax.set_yticklabels([])

            if i != 8: ax.set_xticklabels([])

            if i <= j:
                plt.axis('off')
                continue

            sigma2 = np.median(pairwise_distances(num[i + 1], num[j + 1], metric='euclidean')) ** 2 * 2.0
            mmd2u, mmd2u_null, p_value = two_sample_test(num[i+1], num[j+1], model='MMD',
                                                 kernel_function='rbf', gamma=1.0/sigma2,
                                                 iterations=500, verbose=True, n_jobs=4)

            # print mmd2u
            prob, bins, patches = plt.hist(mmd2u_null, range=[-0.005, 0.1], bins=5, normed=True)
            plt.plot(mmd2u, prob.max() / 10, 'wv', markersize=14, markeredgecolor='k',
                     markeredgewidth=2, label="$MMD^2_u$   p-value < %0.3f" % p_value)
            plt.xlim([-0.01, 0.5])
            if i == 8 and j == 3:
                ax.set_xlabel('$MMD^2_u$ estimate', size=25)
            # plt.legend(loc=1, numpoints=1, prop={'size':15})

    plt.savefig(plot_path + 'MNIST-test.pdf', bbox_inches='tight')


    plt.figure(figsize=(24, 2))

    for i in range(9):

        ax = plt.subplot(1, 9, i+1)

        ax.set_title('%i' % (i + 1), size=25, fontweight="bold")
        if i == 0: ax.set_ylabel('PDF', size=25)
        ax.set_yticklabels([])

        sigma2 = np.median(pairwise_distances(num[i + 1][::2], num[i + 1][1::2], metric='euclidean')) ** 2 * 2.0
        mmd2u, mmd2u_null, p_value = two_sample_test(num[i+1][::2], num[i+1][1::2], model='MMD',
                                                 kernel_function='rbf', gamma=1.0/sigma2,
                                                 iterations=5000, verbose=True, n_jobs=4)

        # print mmd2u
        prob, bins, patches = plt.hist(mmd2u_null, range=[-0.005, 0.01], bins=50, normed=True)
        plt.plot(mmd2u, prob.max() / 10, 'wv', markersize=14, markeredgecolor='k',
                     markeredgewidth=2, label="$MMD^2_u$   p-value < %0.3f" % p_value)
        plt.xlim([-0.005, 0.01])
        if i == 4:
            ax.set_xlabel('$MMD^2_u$ estimate', size=25)
        # plt.legend(loc=1, numpoints=1, prop={'size':15})
        ax.set_xticks([-0.005, 0, 0.005, 0.01 ])
        ax.set_xticklabels([-0.005, '0', 0.005, ' ' ], size=14)

    plt.savefig(plot_path + 'MNIST-test-2.pdf', bbox_inches='tight')


def test_gaussian_vs_lognormal():

    from tatter import two_sample_test
    import numpy as np
    import matplotlib.pyplot as plt
    from sklearn.metrics import pairwise_distances
    import os

    plot_path = './tatter/tests/plots/'
    if not os.path.exists(plot_path):
        os.makedirs(plot_path)

    np.random.seed(100)

    n = 200

    # draw random points from a log-normal and a gaussian distributions
    Y = np.random.lognormal(mean=2, sigma=0.3, size=n)
    X = np.random.normal(loc=np.exp(2.0+0.3**2/2.0), scale=0.3*np.exp(2.0), size=n)

    XX = X[:, np.newaxis]
    YY = Y[:, np.newaxis]

    # compute MMD/KS/KL two sample tests and their null distributions
    sigma2 = np.median(pairwise_distances(XX, YY, metric='euclidean'))**2 * 2.0
    mmd2u, mmd2u_null, p_value = two_sample_test(XX, YY, model='MMD', kernel_function='rbf', gamma=1.0/sigma2,
                                                 iterations=1000, verbose=True, n_jobs=1)

    ks, ks_null, ks_p_value = two_sample_test(XX, YY, model='KS', iterations=1000, verbose=True, n_jobs=1)
    kl, kl_null, kl_p_value = two_sample_test(XX, YY, model='KL', iterations=1000, verbose=True, n_jobs=1)

    # visualize the results
    plt.figure(figsize=(6, 9))
    plt.subplot(311)
    prob, bins, patches = plt.hist(mmd2u_null, range=[-0.005, 0.015], bins=50, normed=True, color='green')
    plt.plot(mmd2u, prob.max()/25, 'wv', markersize=14, markeredgecolor='k',
             markeredgewidth=2, label="$MMD^2_u$   p-value = %0.3f"%p_value)
    plt.xlim([-0.005, 0.020])
    plt.ylabel('PDF', size=20)
    plt.legend(loc=1, numpoints=1, prop={'size':15})

    plt.subplot(312)
    prob, bins, patches = plt.hist(ks_null, range=(0.0, 0.25), bins=50, normed=True, color='orange')
    plt.plot(ks, prob.max()/25, 'wv', markersize=14, markeredgecolor='k',
             markeredgewidth=2, label="$K-S$   p-value = %0.3f"%ks_p_value)
    plt.xlim([0.0, 0.25])
    plt.ylabel('PDF', size=20)
    plt.legend(loc=1, numpoints=1, prop={'size':15})

    plt.subplot(313)
    prob, bins, patches = plt.hist(kl_null, range=(-0.5, 0.7), bins=50, normed=True, color='steelblue')
    plt.plot(kl, prob.max()/25, 'wv', markersize=14, markeredgecolor='k',
             markeredgewidth=2, label="$K-L$    p-value = %0.3f"%kl_p_value)
    plt.xlim([-0.5, 0.7])
    plt.xlabel('Null distribution', size=20)
    plt.ylabel('PDF', size=20)
    plt.legend(loc=1, numpoints=1, prop={'size':15})

    plt.savefig(plot_path + 'log-normal-vs-normal-p-val.pdf', bbox_inches='tight')
    # plt.show()

    plt.figure(figsize=(6, 5))
    prob, bins, patches = plt.hist(X, range=(0.0, 20), bins=20, alpha=0.5,
                                   color='steelblue', normed=True, label='Normal')
    prob, bins, patches = plt.hist(Y, range=(0.0, 20), bins=20, alpha=0.5,
                                   color='lightcoral', normed=True, label='log-Normal')
    plt.legend(loc=1, numpoints=1, prop={'size':16})

    plt.xlabel(r'$x \in \mathcal{X}$', size=18)
    plt.ylabel('PDF', size=18)

    plt.savefig(plot_path + 'log-normal-vs-normal-dist.pdf', bbox_inches='tight')
    # plt.show()




import sys
sys.path.insert(0, sys.path[0]+'/tests')

def main():

    import time
    import warnings
    from tests import \
        test_gaussian_vs_lognormal, \
        RM_clusters_consistency_check, \
        RM_clusters_witness_function, \
        mnist_digits_distance

    warnings.filterwarnings("ignore")

    print("Start ... ")

    start = time.time()

    # test_gaussian_vs_lognormal() ## [passed]
    # RM_clusters_consistency_check() ## [passed]
    # RM_clusters_witness_function() ## [passed]
    mnist_digits_distance() ## [passed]

    end = time.time()
    print("The execution time is %0.2f (s)" % (end - start))

    print("... end. ")


if __name__ == '__main__':
    main()



"""
Two-sample hypothesis test with Maximum Mean Discrepancy (MMD) method.

Introduction:
-------------
TATTER (Two-sAmple TesT EstimatoR) is a tool to perform two-sample hypothesis test.
The two-sample hypothesis test is concerned with whether distributions p(x) and q(x) are
different on the basis of finite samples drawn from each of them. This ubiquitous problem
appears in a legion of applications, ranging from data mining to data analysis and inference.
This implementation can perform the Kolmogorov-Smirnov test (for one-dimensional data only),
Kullback-Leibler divergence, and Maximum Mean Discrepancy (MMD) test. TST comes with
a bootstrap algorithm to estimate the null distribution, and corresponding p-value (statistical
significance of the test value).

The motivation for this approach was to enable scientists to perform two-sample test for
multi-dimensional data sets. This complements the K-S test and K-L divergence which is
commonly used in the astronomy literature.


Quickstart:
-----------
To start using TATTER, simply use "from tatter import two_sample_test" to
access the primary function. The exact requirements for the inputs are
listed in the docstring of the two_sample_test() function further below.
An example for using TATTER looks like this:

    ------------------------------------------------------------
    |  from tatter import two_sample_test                      |
    |                                                          |
    |  test_value, test_null, p_value =                        |
    |           two_sample_test(X, Y,                          |
    |                           model='MMD',                   |
    |                           iterations=1000,               |
    |                           kernel_function='rbf',         |
    |                           gamma=gamma,                   |
    |                           n_jobs=4,                      |
    |                           verbose=True,                  |
    |                           random_state=0)                |
    |                                                          |
    ------------------------------------------------------------


Author:
--------
Arya Farahi, aryaf@umich.edu
Data Science Fellow
University of Michigan -- Ann Arbor


Libraries:
----------
The two-sample test estimator used in this implementation utilizes
'numpy', 'matplotlib', 'sklearn', 'joblib', 'tqdm', and 'pathlib' libraries.


References:
-----------
[1]. A. Farahi, Y. Chen,
     ".", arXiv preprint:

[2]. A. Gretton, B. M. Karsten, R. J. Malte, B. Scholkopf, and A. Smola,
     "A kernel two-sample test."
     Journal of Machine Learning Research 13, no. Mar (2012): 723-773.

[3]. Q. Wang, S. R. Kulkarni, and S. Verdu,
     "Divergence estimation for multidimensional densities via k-nearest-neighbor distances."
     IEEE Transactions on Information Theory 55, no. 5 (2009): 2392-2405.

[4]. W. H. Press, B. P. Flannery, S. A. Teukolsky, and W. T. Vetterling,
     "Numerical recipes." (1989).

"""

from __future__ import division
import numpy as np
from tqdm import tqdm
from scipy import stats
from sklearn.metrics import pairwise_kernels
from KL_estimator import KL_divergence_estimator
from joblib import Parallel, delayed


def MMD2u_estimator(K, m, n):
    """
    Compute the MMD^2_u unbiased statistic.
    This is an implementation of an unbiased MMD^2_u estimator:
    Equation (3) in Gretton et al. Journal of Machine Learning Research 13 (2012) 723-773

    :param K: numpy-array
        the pair-wise kernel matrix
    :param m: int
        dimension of the first data set
    :param n: int
        dimension of the second data set

    :return: float
        an unbiased estimate of MMD^2_u
    """
    K_x = K[:m, :m]
    K_y = K[m:, m:]
    K_xy = K[:m, m:]
    return 1.0 / (m * (m - 1.0)) * (K_x.sum() - K_x.diagonal().sum()) + \
            1.0 / (n * (n - 1.0)) * (K_y.sum() - K_y.diagonal().sum()) - \
            2.0 / (m * n) * K_xy.sum()


def MMD_null_estimator(Kstacked, m, n, rng, **kwargs):
    """
    Compute the MMD^2_u for one bootstrap realization.

    :param Kstacked: numpy-array
        stacked X and Y data vector
    :param m: int
        dimension of the first data set
    :param n: int
        dimension of the second data set
    :param rng: type(np.random.RandomState())
        a numpy random function

    :return: float
        an unbiased estimate of MMD^2_u
    """
    idx = rng.permutation(m + n)
    return MMD2u_estimator(Kstacked[idx, idx[:, None]], m, n)


def KS_null_estimator(XY, m, n, rng, **kwargs):
    """
    Compute the K-S for one bootstrap realization.

    :param XY: numpy-array
        stacked X and Y data vector
    :param m: int
        dimension of the first data set
    :param n: int
        dimension of the second data set
    :param rng: type(np.random.RandomState())
        a numpy random function

    :return: float
        an estimate of K-S
    """
    idx = rng.permutation(m + n)
    K_boot = XY[idx]
    return stats.ks_2samp(K_boot[:m], K_boot[m:])[0]


def KL_null_estimator(XY, m, n, rng, **kwargs):
    """
    Compute the K-L divergence for one bootstrap realization.

    :param XY: numpy-array
        stacked X and Y data vector
    :param m: int
        dimension of the first data set
    :param n: int
        dimension of the second data set
    :param rng: type(np.random.RandomState())
        a numpy random function

    :return: float
        an estimate of K-L divergence
    """
    idx = rng.permutation(m + n)
    K_boot = XY[idx]
    return KL_divergence_estimator(K_boot[:m, :], K_boot[m:, :], **kwargs)


def compute_null_distribution(K, m, n, model='KS', iterations=1000, n_jobs=1,
                              verbose=False, random_state=None, **kwargs):
    """
    Compute the null-distribution of test statistics via a bootstrap algorithm.

    :param K: numpy-array
        the pair-wise kernel matrix, or stacked data sets X and Y
    :param m: int
        dimension of the first data set
    :param n: int
        dimension of the second data set
    :param model: string
        defines the basis model to perform two sample test ['KS', 'KL', 'MMD']
    :param iterations: int
        controls the number of bootstrap realizations
    :param verbose: bool
        controls the verbosity of the model's output.
    :param random_state: type(np.random.RandomState()) or None
        defines the initial random state.
    :param kwargs:
        extra parameters, these are passed to `pairwise_kernels()` as kernel parameters or `KL_divergence_estimator()`
         as the number of k. E.g., if `kernel_two_sample_test(..., kernel_function='rbf', gamma=0.1)`

    :return: numpy-array
        the null distribution
    """

    if model not in ['KS', 'KL', 'MMD']:
        raise ValueError("The Model %s is not implemented, try 'KS', 'KL', or 'MMD'."%model)

    if type(random_state) == type(np.random.RandomState()):
        rng = random_state
    else:
        rng = np.random.RandomState(random_state)

    test_null = np.zeros(iterations)

    if verbose:
        iterations_list = tqdm(range(iterations))
    else:
        iterations_list = range(iterations)

    # compute the null distribution

    # for 1 cpu run the normal code, for more cpu use the Parallel library. This maximize the speed.
    if n_jobs == 1:

        if model == 'MMD':
            for i in iterations_list:
                idx = rng.permutation(m+n)
                K_i = K[idx, idx[:, None]]
                test_null[i] = MMD2u_estimator(K_i, m, n)

        elif model == 'KS':
            for i in iterations_list:
                idx = rng.permutation(m + n)
                K_b = K[idx]
                test_null[i], _ = stats.ks_2samp(K_b[:m], K_b[m:])

        elif model == 'KL':
            for i in iterations_list:
                idx = rng.permutation(m + n)
                K_b = K[idx]
                test_null[i] = KL_divergence_estimator(K_b[:m, :], K_b[m:, :], **kwargs)

    else:
        if model == 'MMD':
            null_estimator = MMD_null_estimator

        elif model == 'KS':
            null_estimator = KS_null_estimator

        elif model == 'KL':
            null_estimator = KL_null_estimator

        test_null = Parallel(n_jobs=n_jobs, prefer="threads")(delayed(null_estimator)(K, m, n, rng, **kwargs)
                                                              for _ in iterations_list)

    return test_null


def two_sample_test(X, Y, model='MMD', kernel_function='rbf', iterations=1000, verbose=False,
                    random_state=None, n_jobs=1, **kwargs):
    """
     This function performs two sample test. The two-sample hypothesis test is concerned with whether distributions $p$
      and $q$ are different on the basis of finite samples drawn from each of them. This ubiquitous problem appears
      in a legion of applications, ranging from data mining to data analysis and inference. This implementation can
      perform the Kolmogorov-Smirnov test (for one-dimensional data only), Kullback-Leibler divergence and MMD.
      The module perform a bootstrap algorithm to estimate the null distribution, and corresponding p-value.

    :param X: numpy-array
        Data, of size MxD [M is the number of data points, D is the features dimension]
    :param Y: numpy-array
        Data, of size NxD [N is the number of data points, D is the features dimension]
    :param model: string
        defines the basis model to perform two sample test ['KS', 'KL', 'MMD']
    :param kernel_function: string
        defines the kernel function, only used for the MMD.
        For the list of implemented kernel please consult with https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.kernel_metrics.html#sklearn.metrics.pairwise.kernel_metrics
    :param iterations: int
        controls the number of bootstrap realizations
    :param verbose: bool
        controls the verbosity of the model's output.
    :param random_state: type(np.random.RandomState()) or None
        defines the initial random state.
    :param n_jobs: int
        number of jobs to run in parallel.
    :param kwargs:
        extra parameters, these are passed to `pairwise_kernels()` as kernel parameters or `KL_divergence_estimator()`
         as the number of k. E.g., if `kernel_two_sample_test(..., kernel_function='rbf', gamma=0.1)`

    :return: tuple of size 3
        float: the test value,
        numpy-array: a null distribution via bootstraps,
        float: estimated p-value
    """

    if model not in ['KS', 'KL', 'MMD']:
        raise ValueError("The Model '%s' is not implemented, try 'KS', 'KL', or 'MMD'."%model)

    if X.shape[1] != Y.shape[1]:
        raise ValueError("Incompatible dimension for X and Y matrices. X and Y should have the same feature dimension,"
                         ": X.shape[1] == %i while Y.shape[1] == %i."%(X.shape[1], Y.shape[1]))

    if model == 'KS' and X.shape[1] > 1:
        raise ValueError("The KS test can handle only one dimensional data,"
                         ": X.shape[1] == %i and Y.shape[1] == %i."%(X.shape[1], Y.shape[1]))

    if not (isinstance(iterations, int) and iterations > 1):
        raise ValueError('iterations has incorrect type or less than 2.')

    if not (isinstance(n_jobs, int) and n_jobs > 0):
        raise ValueError('n_jobs is incorrect type or <1. n_jobs:%s'%n_jobs)

    m = len(X)
    n = len(Y)

    # p-value's resolution
    resolution = 1.0/iterations

    # compute the test statistics according to the input model
    if model == 'MMD':
        XY = np.vstack([X, Y])
        K = pairwise_kernels(XY, metric=kernel_function, **kwargs)
        test_value = MMD2u_estimator(K, m, n)

    elif model == 'KS':
        K = np.concatenate((X.T[0], Y.T[0]))
        test_value, _ = stats.ks_2samp(X.T[0], Y.T[0])

    elif model == 'KL':
        K = np.vstack([X, Y])
        test_value = KL_divergence_estimator(X, Y, **kwargs)

    if verbose:
        print("test value = %s"%test_value)
        print("Computing the null distribution.")

    # compute the null distribution via a bootstrap algorithm
    test_null = compute_null_distribution(K, m, n, model=model, iterations=iterations, verbose=verbose,
                                          n_jobs=n_jobs, random_state=random_state, **kwargs)

    # compute the p-value, if less then the resolution set it to the resolution
    p_value = max(resolution, resolution*(test_null > test_value).sum())

    if verbose:
        if p_value == resolution:
            print("p-value < %s \t (resolution : %s)" % (p_value, resolution))
        else:
            print("p-value ~= %s \t (resolution : %s)" % (p_value, resolution))

    return test_value, test_null, p_value


def test_statistics(X, Y, model='MMD', kernel_function='rbf', **kwargs):
    """
     This function performs a test statistics and return a test value. This implementation can perform
     the Kolmogorov-Smirnov test (for one-dimensional data only), Kullback-Leibler divergence and MMD.

    :param X: numpy-array
        Data, of size MxD [M is the number of data points, D is the features dimension]
    :param Y: numpy-array
        Data, of size NxD [N is the number of data points, D is the features dimension]
    :param model: string
        defines the basis model to perform two sample test ['KS', 'KL', 'MMD']
    :param kernel_function: string
        defines the kernel function, only used for the MMD.
        For the list of implemented kernel please consult with https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.kernel_metrics.html#sklearn.metrics.pairwise.kernel_metrics
    :param kwargs:
        extra parameters, these are passed to `pairwise_kernels()` as kernel parameters or `KL_divergence_estimator()`
        as the number of k. E.g., if `kernel_two_sample_test(..., kernel_function='rbf', gamma=0.1)`

    :return: float
        the test value
    """

    if model not in ['KS', 'KL', 'MMD']:
        raise ValueError("The Model '%s' is not implemented, try 'KS', 'KL', or 'MMD'." % model)

    if X.shape[1] != Y.shape[1]:
        raise ValueError("Incompatible dimension for X and Y matrices. X and Y should have the same feature dimension,"
                         ": X.shape[1] == %i while Y.shape[1] == %i." % (X.shape[1], Y.shape[1]))

    if model == 'KS' and X.shape[1] > 1:
        raise ValueError("The KS test can handle only one dimensional data,"
                         ": X.shape[1] == %i and Y.shape[1] == %i." % (X.shape[1], Y.shape[1]))

    m = len(X)
    n = len(Y)

    # compute the test statistics according to the input model
    if model == 'MMD':
        XY = np.vstack([X, Y])
        K = pairwise_kernels(XY, metric=kernel_function, **kwargs)
        test_value = MMD2u_estimator(K, m, n)

    elif model == 'KS':
        test_value, _ = stats.ks_2samp(X.T[0], Y.T[0])

    elif model == 'KL':
        test_value = KL_divergence_estimator(X, Y, **kwargs)

    return test_value


def witness_function(X, Y, grid, kernel_function='rbf', **kwargs):
    """
     This function computes the witness function. For the definition of the witness function see page 729
     in the "A Kernel Two-Sample Test" by Gretton et al. (2012)

    :param X: numpy-array
        Data, of size MxD [M is the number of data points, D is the features dimension]
    :param Y: numpy-array
        Data, of size NxD [N is the number of data points, D is the features dimension]
    :param gird: numpy-array
        Defines a grid for which the witness function is computed. It has the size PxD
        where P is the number of grid points, D is the features dimension
    :param kernel_function: string
        defines the kernel function, only used for the MMD.
        For the list of implemented kernel please consult with https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.kernel_metrics.html#sklearn.metrics.pairwise.kernel_metrics
    :param kwargs:
        extra parameters, these are passed to `pairwise_kernels()` as kernel parameters.
        E.g., if `kernel_two_sample_test(..., kernel_function='rbf', gamma=0.1)`

    :return: numpy-array
        witness function
    """

    if X.shape[1] != Y.shape[1]:
        raise ValueError("Incompatible dimension for X and Y matrices. X and Y should have the same feature dimension,"
                         ": X.shape[1] == %i while Y.shape[1] == %i." % (X.shape[1], Y.shape[1]))

    if X.shape[1] != grid.shape[1]:
        raise ValueError("Incompatible dimension for data and grid matrices. data and grid should have the same feature dimension,"
                         ": data.shape[1] == %i while grid.shape[1] == %i." % (X.shape[1], grid.shape[1]))

    # data and grid size
    m = len(X)
    n = len(Y)

    # compute pairwise kernels
    K_xg = pairwise_kernels(X, grid, metric=kernel_function, **kwargs)
    K_yg = pairwise_kernels(Y, grid, metric=kernel_function, **kwargs)

    return (np.sum(K_xg, axis=0) / m) - (np.sum(K_yg, axis=0) / n)




def if_RM_clusters_exists():

    from pathlib import Path
    import os

    data_path = "./tatter/tests/data/"

    if not os.path.exists(data_path):
        os.makedirs(data_path)

    # if cluster catalogs does noe exists download them
    for fname in ["redmapper_sva1_public_v6.3_catalog.fits", "redmapper_dr8_public_v6.3_catalog.fits"]:

        fname_file = Path(data_path + fname)

        if not fname_file.is_file():
            import wget, gzip, shutil
            url = 'http://risa.stanford.edu/redmapper/v6.3/%s.gz' % fname
            _ = wget.download(url, out=data_path[2:])
            with gzip.open(data_path + fname + '.gz', 'rb') as f_in:
                with open(data_path + fname, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            os.remove(data_path + fname + '.gz')


def RM_clusters_witness_function():

    import numpy as np
    from astropy.io import fits
    import matplotlib.pyplot as plt
    from sklearn.metrics import pairwise_distances
    from tatter import two_sample_test, test_statistics, witness_function
    import os

    data_path = "./tatter/tests/data/"
    plot_path = './tatter/tests/plots/'
    if not os.path.exists(plot_path):
        os.makedirs(plot_path)

    # if cluster catalogs does noe exists download them
    if_RM_clusters_exists()

    plt.figure(figsize=(12, 12))

    for i in range(4):

        zmin = 0.2 + 0.05*i
        zmax = 0.2 + 0.05*(i+1)

        # load the datasets
        RM_SDSS = fits.open(data_path+'redmapper_dr8_public_v6.3_catalog.fits')[1].data
        RM_SV = fits.open(data_path+'redmapper_sva1_public_v6.3_catalog.fits')[1].data

        # apply a redshift and richness cut
        RM_SDSS = RM_SDSS[RM_SDSS.Z_LAMBDA >= zmin]
        RM_SDSS = RM_SDSS[RM_SDSS.Z_LAMBDA < zmax]
        RM_SDSS = RM_SDSS[RM_SDSS.LAMBDA > 20]

        RM_SV = RM_SV[RM_SV.Z_LAMBDA >= zmin]
        RM_SV = RM_SV[RM_SV.Z_LAMBDA < zmax]
        RM_SV = RM_SV[RM_SV.LAMBDA > 20]

        # map richness into a log space
        RM_SDSS.LAMBDA = np.log(RM_SDSS.LAMBDA)
        RM_SV.LAMBDA = np.log(RM_SV.LAMBDA)

        # make a numpy array
        SDSS = np.array(RM_SDSS.LAMBDA)[:, np.newaxis]
        SV = np.array(RM_SV.LAMBDA)[:, np.newaxis]

        # compute the hyper-parameter
        sigma2 = np.median(pairwise_distances(SDSS, SV, metric='euclidean')) ** 2 * 2.0

        # estimate the null distribution and p-value of MMD^2_u test.
        grid = np.linspace(np.log(20), np.log(100), 1001)[:, np.newaxis]
        wf = witness_function(SV, SDSS, grid, kernel_function='rbf', gamma=1.0 / sigma2)

        ax = plt.subplot(2, 2, i+1)
        plt.plot(np.exp(grid), wf, color='black', lw=2.0)

        plt.xlim([20, 100])
        plt.title('$ %0.2f \leq z < %0.2f$'%(zmin, zmax), size=14)

        if i%2 == 0: ax.set_ylabel('witness function', size=22)
        if i > 1: ax.set_xlabel(r'$\lambda_{\rm RM}$', size=22)

    plt.savefig(plot_path + 'SDSS-vs-SV-witness-function.pdf', bbox_inches='tight')


def RM_clusters_consistency_check():

    import numpy as np
    from astropy.io import fits
    from tatter import two_sample_test
    import matplotlib.pyplot as plt
    from sklearn.metrics import pairwise_distances
    import os

    plt.figure(figsize=(12, 12))

    data_path = "./tatter/tests/data/"
    plot_path = './tatter/tests/plots/'
    if not os.path.exists(plot_path):
        os.makedirs(plot_path)

    # if cluster catalogs does noe exists download them
    if_RM_clusters_exists()

    for i in range(4):

        zmin = 0.2 + 0.05*i
        zmax = 0.2 + 0.05*(i+1)

        # load the datasets
        RM_SDSS = fits.open(data_path+'redmapper_dr8_public_v6.3_catalog.fits')[1].data
        RM_SV = fits.open(data_path+'redmapper_sva1_public_v6.3_catalog.fits')[1].data

        # apply a redshift cut
        RM_SDSS = RM_SDSS[RM_SDSS.Z_LAMBDA >= zmin]
        RM_SDSS = RM_SDSS[RM_SDSS.Z_LAMBDA < zmax]

        RM_SV = RM_SV[RM_SV.Z_LAMBDA >= zmin]
        RM_SV = RM_SV[RM_SV.Z_LAMBDA < zmax]

        # map richness into a log space
        RM_SDSS.LAMBDA = np.log(RM_SDSS.LAMBDA)
        RM_SV.LAMBDA = np.log(RM_SV.LAMBDA)

        # normalize richness -- (x - x.median) / x.std
        std = np.std(RM_SDSS.LAMBDA); med = np.median(RM_SDSS.LAMBDA)

        RM_SV.LAMBDA = (RM_SV.LAMBDA - med) / std
        RM_SDSS.LAMBDA = (RM_SDSS.LAMBDA - med) / std

        # normalize redshift -- (x - x.median) / x.std
        std = np.std(RM_SDSS.Z_LAMBDA); med = np.median(RM_SDSS.Z_LAMBDA)

        RM_SV.Z_LAMBDA = (RM_SV.Z_LAMBDA - med) / std
        RM_SDSS.Z_LAMBDA = (RM_SDSS.Z_LAMBDA - med) / std

        # make a numpy array
        SDSS = np.array([RM_SDSS.LAMBDA, RM_SDSS.Z_LAMBDA]).T
        SV = np.array([RM_SV.LAMBDA, RM_SV.Z_LAMBDA]).T

        # compute the hyper-parameter
        sigma2 = np.median(pairwise_distances(SDSS, SV, metric='euclidean')) ** 2 * 2.0
        # estimate the null distribution and p-value of MMD^2_u test.
        mmd2u, mmd2u_null, p_value = two_sample_test(SDSS, SV, model='MMD',
                                                     kernel_function='rbf', gamma=1.0/sigma2,
                                                     iterations=500, verbose=True, n_jobs=4)

        ax = plt.subplot(2, 2, i+1)
        prob, bins, patches = ax.hist(mmd2u_null, range=(-0.01, 0.03), bins=100, normed=True)
        if p_value <= 0.001:
            plt.plot(mmd2u, prob.max() / 25, 'wv', markersize=22, markeredgecolor='k',
                     markeredgewidth=2, label='$MMD^2_u$ ($p$-value < %0.3f)'%p_value)
        else:
            plt.plot(mmd2u, prob.max() / 25, 'wv', markersize=22, markeredgecolor='k',
                     markeredgewidth=2, label='$MMD^2_u$ ($p$-value = %0.3f)'%p_value)
        plt.xlim([-0.01, 0.03])
        plt.title('$ %0.2f \leq z < %0.2f$'%(zmin, zmax), size=14)

        if i%2 == 0: ax.set_ylabel('$PDF$', size=22)
        if i > 1: ax.set_xlabel('$MMD^2_u$', size=22)
        plt.legend(numpoints=1, prop={'size':14})

    plt.savefig(plot_path + 'SDSS-vs-SV.pdf', bbox_inches='tight')




"""
Posterior Predictive check with Maximum Mean Discrepancy (MMD).

Introduction:
-------------
TATTER (Two-sAmple TesT EstimatoR) is equipped with a posterior predictive check that allows
the users to study the goodness of a data generative model vs. an observed data. This module employ
MMD, KL, KS test to compute the distance a set of simulated dataset.


Quickstart:
-----------
To start using TATTER, simply use "from tatter import posterior_predictive_check"
to access the posterior predictive check function. The exact requirements
for the inputs are listed in the docstring of the  posterior_predictive_check()
function further below. An example of this function looks like this:

    -----------------------------------------------------------------
    |  from tatter import posterior_predictive_check                |
    |                                                               |
    |  test_value, test_null, p_value =                             |
    |          posterior_predictive_check(X, sims,                  |
    |                                     model='MMD',              |
    |                                     kernel_function='rbf',    |
    |                                     gamma=gamma,              |
    |                                     n_jobs=4,                 |
    |                                     verbose=True,             |
    |                                     random_state=0)           |
    |                                                               |
    -----------------------------------------------------------------


Author:
--------
Arya Farahi, aryaf@umich.edu
Data Science Fellow
University of Michigan -- Ann Arbor


Libraries:
----------
The two-sample test estimator used in this implementation utilizes
'numpy', 'matplotlib', 'sklearn', 'joblib', 'tqdm', and 'pathlib' libraries.


References:
-----------
[1]. A. Gelman, X. L. Meng, and H. Stern,
    "Posterior predictive assessment of model fitness via realized discrepancies."
    Statistica sinica (1996): 733-760.
"""

from __future__ import division
import numpy as np
from tqdm import tqdm
from scipy import stats
from sklearn.metrics import pairwise_kernels
from .KL_estimator import KL_divergence_estimator
from .TST_estimators import MMD2u_estimator, MMD_null_estimator,\
    KS_null_estimator, KL_divergence_estimator, KL_null_estimator
from joblib import Parallel, delayed

def posterior_predictive_check(X, sims, model='MMD', kernel_function='rbf', verbose=False,
                               random_state=None, n_jobs=1, **kwargs):
    """
     This function performs a posterior predictive check. The posterior predictive check is a
      Bayesian counterpart of the classical tests for goodness-of-the-fit. It often can be used in
      judging the fit of a single Bayesian model to the observed data. The posterior predictive
      check require a test statistics. In our implementation, the test sample here can be one of
      the following tests: Kolmogorov-Smirnov, Kullback-Leibler divergence, or MMD.
      We note that the Kolmogorov-Smirnov test is defined for only one-dimensional data.
      The module perform a bootstrap algorithm to estimate the null distribution, and corresponding p-value.

    :param X: numpy-array
        Observed data, of size M x D [ is the number of data points, D is the features dimension]
    :param sims: list of numpy-array
        A list of a set of simulated data, of size N x D
        [Nsim is the number of simulatins, N is the number of data points, D is the features dimension]
    :param model: string
        defines the basis model to perform two sample test ['KS', 'KL', 'MMD']
    :param kernel_function: string
        defines the kernel function, only used for the MMD.
        For the list of implemented kernel please consult with https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.kernel_metrics.html#sklearn.metrics.pairwise.kernel_metrics
    :param verbose: bool
        controls the verbosity of the model's output.
    :param random_state: type(np.random.RandomState()) or None
        defines the initial random state.
    :param n_jobs: int [not implimented yet]
        number of jobs to run in parallel.
    :param kwargs:
        extra parameters, these are passed to `pairwise_kernels()` as kernel parameters or `KL_divergence_estimator()`
         as the number of k. E.g., if `kernel_two_sample_test(..., kernel_function='rbf', gamma=0.1)`

    :return: tuple of size 3
        float: the test value,
        numpy-array: a null distribution via bootstraps,
        float: estimated p-value
    """

    if model not in ['KS', 'KL', 'MMD']:
        raise ValueError("The Model '%s' is not implemented, try 'KS', 'KL', or 'MMD'."%model)

    # if len(X.shape) != 2 or len(sim.shape) != 3:
    #    raise ValueError("Incompatible shape for X and sim matrices. X and sim should have a shape of length 2 and 3 respectively,"
    #                     ": len(X.shape) == %i while len(sim.shape) == %i."%(len(X.shape), len(sim.shape)))

    # if X.shape[1] != sim.shape[1]:
    #    raise ValueError("Incompatible dimension for X and Y matrices. X and Y should have the same feature dimension,"
    #                     ": X.shape[1] == %i while Y.shape[1] == %i."%(X.shape[1], Y.shape[1]))

    if model == 'KS' and X.shape[1] > 1:
        raise ValueError("The KS test can handle only one dimensional data,"
                         ": X.shape[1] == %i and Y.shape[1] == %i."%(X.shape[1], Y.shape[1]))

    if not (isinstance(n_jobs, int) and n_jobs > 0):
        raise ValueError('n_jobs is incorrect type or <1. n_jobs:%s'%n_jobs)

    # define the random state
    if type(random_state) == type(np.random.RandomState()):
        rng = random_state
    else:
        rng = np.random.RandomState(random_state)

    m = len(X)

    # p-value's resolution
    resolution = 1.0/len(sims)

    test_value = []
    null_value = []

    # compute the test statistics according to the input model
    # compute the null distribution via a bootstrap algorithm
    if model == 'MMD':

        for Y in sims:

            n = len(Y)
            XY = np.vstack([X, Y])
            K = pairwise_kernels(XY, metric=kernel_function, **kwargs)

            test_value += [MMD2u_estimator(K, m, n)]
            # nv = np.mean([MMD_null_estimator(K, m, n, rng) for i in range(20)])
            null_value += [MMD_null_estimator(K, m, n, rng)]

    elif model == 'KS':

        for Y in sims:

            n = len(Y)
            K = np.concatenate((X.T[0], Y.T[0]))

            test_value += [stats.ks_2samp(X.T[0], Y.T[0])[0]]
            null_value += [KS_null_estimator(K, m, n, rng)]

    elif model == 'KL':

        for Y in sims:
            n = len(Y)
            K = np.vstack([X, Y])

            test_value += [KL_divergence_estimator(X, Y, **kwargs)]
            null_value += [KL_null_estimator(K, m, n, rng)]

    test_value = np.array(test_value)
    null_value = np.array(null_value)

    if verbose:
        print("test value = %s"%test_value)
        print("Computing the null distribution.")

    # compute the p-value, if less then the resolution set it to the resolution
    p_value = max(resolution, resolution*(null_value > test_value).sum())

    if verbose:
        if p_value == resolution:
            print("p-value < %s \t (resolution : %s)" % (p_value, resolution))
        else:
            print("p-value ~= %s \t (resolution : %s)" % (p_value, resolution))

    return test_value, null_value, p_value


def test_statistics(X, Y, model='MMD', kernel_function='rbf', **kwargs):
    """
     This function performs a test statistics and return a test value. This implementation can perform
     the Kolmogorov-Smirnov test (for one-dimensional data only), Kullback-Leibler divergence and MMD.

    :param X: numpy-array
        Data, of size MxD [M is the number of data points, D is the features dimension]
    :param Y: numpy-array
        Data, of size NxD [N is the number of data points, D is the features dimension]
    :param model: string
        defines the basis model to perform two sample test ['KS', 'KL', 'MMD']
    :param kernel_function: string
        defines the kernel function, only used for the MMD.
        For the list of implemented kernel please consult with https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.kernel_metrics.html#sklearn.metrics.pairwise.kernel_metrics
    :param kwargs:
        extra parameters, these are passed to `pairwise_kernels()` as kernel parameters or `KL_divergence_estimator()`
        as the number of k. E.g., if `kernel_two_sample_test(..., kernel_function='rbf', gamma=0.1)`

    :return: float
        the test value
    """

    if model not in ['KS', 'KL', 'MMD']:
        raise ValueError("The Model '%s' is not implemented, try 'KS', 'KL', or 'MMD'." % model)

    if X.shape[1] != Y.shape[1]:
        raise ValueError("Incompatible dimension for X and Y matrices. X and Y should have the same feature dimension,"
                         ": X.shape[1] == %i while Y.shape[1] == %i." % (X.shape[1], Y.shape[1]))

    if model == 'KS' and X.shape[1] > 1:
        raise ValueError("The KS test can handle only one dimensional data,"
                         ": X.shape[1] == %i and Y.shape[1] == %i." % (X.shape[1], Y.shape[1]))

    m = len(X)
    n = len(Y)

    # compute the test statistics according to the input model
    if model == 'MMD':
        XY = np.vstack([X, Y])
        K = pairwise_kernels(XY, metric=kernel_function, **kwargs)
        test_value = MMD2u_estimator(K, m, n)

    elif model == 'KS':
        test_value, _ = stats.ks_2samp(X.T[0], Y.T[0])

    elif model == 'KL':
        test_value = KL_divergence_estimator(X, Y, **kwargs)

    return test_value
